from __future__ import division
from __future__ import absolute_import
from .iapws import Water

# Water at triple point
pt = 612
Tt = 273.16


class MoistAir(object):
    """
    Class that models a moist gas.  The computations in this class are
    a bit tricky because the enthalpy computations for steam far from
    atmospheric pressure using the bcura data have severe
    deviations. Considering water an ideal gas is a too strong
    assumption. This means that water data is computed using the IAPWS
    data for water an steam.

    The trickiest part comes with the fact that the enthalpy reference
    for the IAPWS tables is the water's triple point when the enthalpy
    reference for the bcura tables is the absolute zero.

    The IAPWS reference is the leading one.

    MoistAir does not inherit from Mixture because of this.
    """

    def __init__(self, gas):
        self.moist = Water()

        (self.water, self.qwater) = gas['H2O']
        if self.water is None:
            raise ValueError("No water in this gas")

        gas.delete('H2O')
        self.gas = gas

        # amount of gas (in case it is not 100-qwater)
        self.qgas = 0
        for e in self.gas:
            self.qgas += e[1]

        self.w = self.water.mm / self.gas.mm * self.qwater / self.qgas

    def phi(self, p, T):
        """
        Relative moisture given pressure and temperature.
        """
        ya = self.qgas / (self.qgas + self.qwater)
        return self.gas.mm * ya * p * self.w / (self.water.mm *
                                                self.moist.psat(T))

    def wet_bulb_T(self, p):
        """
        Wet bulb temperature for a given pressure
        """
        yw = self.qwater / (self.qgas + self.qwater)
        return self.moist.Tsat(yw * p)

    def h(self, p, T):
        raise NotImplementedError

    def __repr__(self):
        return "<Moist Gas>:\n  Gas:\n" + self.gas.__repr__()

    def __unicode__(self):
        return u"<Moist Gas>:\n  Gas:\n" + self.gas.__unicode__()



from __future__ import division
from __future__ import absolute_import
from .burcat import Mixture
from scipy import optimize
from copy import deepcopy

supported = 'C H N O'.split()

# TODO: Check that adiabatic flame temperature is ok


def balance(fuel, am, phi):
    """
    Function that balances the combustion equation given any simple
    fuel element

    :fuel:

      Formula of a Burcat's database fuel as a string. The fuel can
      only contain C, H, O and N.  Other atoms will be ignored.

    :phi:

      Equivalence ratio for air.
    """
    reactants = {'fuel': am, 'N2': 0, 'O2': 0}
    products = {'N2': 0, 'CO2': 0, 'H2O': 0, 'O2': 0}
    atoms = {'C': 0, 'H': 0, 'N': 0, 'O': 0}

    lamb = 1 / phi
    for element in fuel.elements:
        if element[0] in supported:
            atoms[element[0]] = element[1]

    k = lamb * (atoms['C'] + atoms['H'] / 4 + atoms['O'] / 2)
    reactants['O2'] = am * k
    reactants['N2'] = am * 3.76 * k

    products['N2'] = am * (atoms['N'] + 3.76 * k)
    products['CO2'] = am * (atoms['C'])
    products['H2O'] = am * (atoms['H'] / 2)
    products['O2'] = am * ((lamb - 1) * (k / lamb))

    return (reactants, products)


def balance_mix(fuels, phi):
    """
    function that balances the combustion equation given a mix of
    fuels.

    :fuels:

      type Mixture. Simple fuels formed only by C, H, O and N and the
      amount of each one

    :phi:

      Equivalence ratio for air.
    """
    reactants = {'N2': 0, 'O2': 0}
    products = {'N2': 0, 'CO2': 0, 'H2O': 0, 'O2': 0}
    atoms = {'C': 0, 'H': 0, 'N': 0, 'O': 0}

    for fuel in fuels:
        (dreac, dprod) = balance(fuel[0], fuel[1], phi)
        reactants['O2'] += dreac['O2']
        reactants['N2'] += dreac['N2']
        products['N2'] += dprod['N2']
        products['CO2'] += dprod['CO2']
        products['H2O'] += dprod['H2O']
        products['O2'] += dprod['O2']

    return (reactants, products)


class SimpleCombustor(object):
    """
    This class models a simple combustor that uses fuel as a reductor
    and air as a single oxidizer. The combustion is complete, no CO
    nor radicals are formed.

    It only supports fuels with C, N, H and O.  If you put a more
    complicated fuel it will ignore the rest of atoms to balance the
    reaction.

    """

    def __init__(self, fuel, phi, db):
        """
        :fuel:

          The formula of the fuel

        :phi:

          Equivalence ratio

        :db:

          Element database.  It is passed as argument because it has
          to be allocated only once.

        >>> from .burcat import Elementdb
        >>> db = Elementdb()
        >>> methane = db.getelementdata("CH4   RRHO")
        >>> combustor = SimpleCombustor(methane,1,db)
        >>> print(round(combustor.products.cp, 6))
        1112.222587
        >>> print(round(combustor.heat_of_comb(298.15), 2))
        50027136.34
        >>> print(round(combustor.adiabatic_flame_temp(298.15)[0], 6))
        2128.632842
        """

        # REACTANTS
        self.reactants = Mixture()
        self.products = Mixture()

        oxygen = db.getelementdata("O2 REF ELEMENT")
        nitrogen = db.getelementdata("N2  REF ELEMENT")
        carbondiox = db.getelementdata("CO2")
        water = db.getelementdata("H2O")

        (rdict, pdict) = balance(fuel, 1, phi)

        self.reactants.add(fuel, rdict['fuel'])
        self.reactants.add(oxygen, rdict['O2'])
        self.reactants.add(nitrogen, rdict['N2'])

        self.products.add(nitrogen, pdict['N2'])
        self.products.add(carbondiox, pdict['CO2'])
        self.products.add(water, pdict['H2O'])
        self.products.add(oxygen, pdict['O2'])

    def heat_of_comb(self, T):
        """
        Calculates the heat of combustion per kg of fuel. Checked ok
        """
        hreac = float(0)
        for reac in self.reactants:
            hreac += reac[0].ho(T) * reac[1]

        hprod = float(0)
        for prod in self.products:
            hprod += prod[0].ho(T) * prod[1]

        return float(hreac - hprod) / self.reactants[0][0].mm

    @property
    def lower_heating_value(self):
        return self.heat_of_comb(423.15)

    def adiabatic_flame_temp(self, T):
        """
        This is the adiabatic flame temp for the given mixtures of
        reactants and products.  If you want the true adiabatic flame
        temperature remember to set the equivalence ratio to 1.
        Otherwise you will always get lower temperatures.
        """
        dh = self.heat_of_comb(T) * self.reactants[0][0].mm

        f = \
            lambda Tg: self.products[0][1] * self.products[0][0].cpo(Tg) +\
            self.products[1][1] * self.products[1][0].cpo(Tg) +\
            self.products[2][1] * self.products[2][0].cpo(Tg) +\
            self.products[3][1] * self.products[3][0].cpo(Tg) - dh / (Tg - T)

        return optimize.fsolve(f, 1000)


class Combustor(object):
    """
    Combustor that is able to characterize the combustion of a mixture
    of fuels
    """

    def __init__(self, fuels, phi, db):
        self.fuels = deepcopy(fuels)
        self.reactants = fuels
        self.products = Mixture()

        oxygen = db.getelementdata("O2 REF ELEMENT")
        nitrogen = db.getelementdata("N2  REF ELEMENT")
        carbondiox = db.getelementdata("CO2")
        water = db.getelementdata("H2O")

        (dreac, dprod) = balance_mix(fuels, phi)

        self.reactants.add(oxygen, dreac['O2'])
        self.reactants.add(nitrogen, dreac['N2'])

        self.products.add(nitrogen, dprod['N2'])
        self.products.add(carbondiox, dprod['CO2'])
        self.products.add(water, dprod['H2O'])
        self.products.add(oxygen, dprod['O2'])

    def heat_of_comb(self, T):
        """
        Calculates the heat of combustion per kg of fuel. Checked ok
        """
        hreac = float(0)
        for reac in self.reactants:
            hreac += reac[0].ho(T) * reac[1]

        hprod = float(0)
        for prod in self.products:
            hprod += prod[0].ho(T) * prod[1]

        return float(hreac - hprod) / self.fuels.mm

    @property
    def lower_heating_value(self):
        return self.heat_of_comb(423.15)

    def adiabatic_flame_temp(self, T):
        """
        This is the adiabatic flame temp for the given mixtures of
        reactants and products.  If you want the true adiabatic flame
        temperature remember to set the equivalence ratio to 1.
        Otherwise you will always get lower temperatures.
        """
        dh = self.heat_of_comb(T) * self.fuels.mm

        f = \
            lambda Tg: self.products[0][1] * self.products[0][0].cpo(Tg) +\
            self.products[1][1] * self.products[1][0].cpo(Tg) +\
            self.products[2][1] * self.products[2][0].cpo(Tg) +\
            self.products[3][1] * self.products[3][0].cpo(Tg) - dh / (Tg - T)

        return optimize.fsolve(f, 1000)



"""
This module gets thermodynamic datat from the JANAF database.
Files are downloaded from the NIST servers as needed and then cached locally.

Zack Gainsforth

Funding by NASA
"""

from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import os

try:
    # Python 3
    import urllib.request as urllib2
except ImportError:
    # Python 2
    import urllib2

try:
    # Python 2
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO

# Universal gas constant R
R = 8.314472


class JanafPhase(object):
    """
    Class which is created by Janafdb for a specific phase.

    It reads in the JANAF data file and produces functions which interpolate
    the thermodynamic constants.

    >>> db = Janafdb()
    >>> p = db.getphasedata(formula='O2Ti', name='Rutile', phase='cr')
    >>> print(p.cp([500, 550, 1800]))
    [ 67.203  68.567  78.283]
    >>> print(p.S([500, 550, 1800]))
    [  82.201    88.4565  176.876 ]
    >>> print(p.gef([500, 550, 1800]))
    [  57.077   59.704  115.753]
    >>> print(p.hef([500, 550, 1800]))
    [  12.562    15.9955  110.022 ]
    >>> print(p.DeltaH([500, 550, 1800]))
    [-943670.  -943229.5 -936679. ]
    >>> print(p.DeltaG([500, 550, 1800]))
    [-852157.  -843046.5 -621013. ]
    >>> print(p.logKf([500, 550, 1800]))
    [ 89.024   80.8125  18.021 ]
    >>> print(p.cp(50000))
    Traceback (most recent call last):
        ...
    ValueError: A value in x_new is above the interpolation range.
    """

    def __init__(self, rawdata_text):
        # Store the raw data text file from NIST.
        self.rawdata_text = rawdata_text

        # Read the text file into a DataFrame.
        data = pd.read_csv(
            StringIO(self.rawdata_text),
            skiprows=2,
            header=None,
            delimiter='[\t\s]*',
            engine='python')
        data.columns = ['T', 'Cp', 'S', '[G-H(Tr)]/T', 'H-H(Tr)', 'Delta_fH',
                        'Delta_fG', 'log(Kf)']
        self.rawdata = data

        # Sometimes the JANAF files have funky stuff written in them. (Old
        # school text format...)
        # Clean it up.
        for c in data.columns:
            # We only need to polish up columns that aren't floating point
            # numbers.
            if np.issubdtype(data.dtypes[c], np.floating):
                continue
            # Change INFINITE to inf
            data.loc[data[c] == 'INFINITE', c]
            # Anything else becomes a nan.
            # Convert to floats.
            data[c] = pd.to_numeric(data[c], errors='coerce')

        # Convert all units to Joules.
        data['Delta_fH'] *= 1000
        data['Delta_fG'] *= 1000

        # Now make interpolatable functions for each of these.
        self.cp = interp1d(self.rawdata['T'], self.rawdata['Cp'])
        self.S = interp1d(self.rawdata['T'], self.rawdata['S'])
        self.gef = interp1d(self.rawdata['T'], self.rawdata['[G-H(Tr)]/T'])
        self.hef = interp1d(self.rawdata['T'], self.rawdata['H-H(Tr)'])
        self.DeltaH = interp1d(self.rawdata['T'], self.rawdata['Delta_fH'])
        self.DeltaG = interp1d(self.rawdata['T'], self.rawdata['Delta_fG'])
        self.logKf = interp1d(self.rawdata['T'], self.rawdata['log(Kf)'])

        # TODO Deal well with crystal<->liquid transitions which have a below
        # and above value for Cp, S, etc.


class Janafdb(object):
    """
    Class that reads the NIST JANAF tables for thermodynamic data.

    Data is initially read from the web servers, and then cached.
    """
    VALIDPHASETYPES = ['cr', 'l', 'cr,l', 'g', 'ref', 'cd', 'fl', 'am', 'vit',
                       'mon', 'pol', 'sln', 'aq', 'sat']
    JANAF_URL = "http://kinetics.nist.gov/janaf/html/%s.txt"

    def __init__(self):
        """
        We have an index file which can be used to build the url for all phases
        on the NIST site.
        """

        # Read the index file which tells us the filenames for all the phases
        # in the JANAF database.
        dirname = os.path.dirname(__file__)
        janaf_index = os.path.join(dirname, 'JANAF_index.txt')
        self.db = pd.read_csv(janaf_index, delimiter='|')
        # Name the columns and trim whitespace off the text fields.
        self.db.columns = ['formula', 'name', 'phase', 'filename']
        self.db["formula"] = self.db["formula"].map(str.strip)
        self.db["name"] = self.db["name"].map(str.strip)
        self.db["phase"] = self.db["phase"].map(str.strip)
        self.db["filename"] = self.db["filename"].map(str.strip)

        # Make sure that the directory for cached JANAF files exists.
        self.JANAF_cachedir = os.path.join(dirname, 'JANAF_Cache')
        if not os.path.exists(self.JANAF_cachedir):
            os.mkdir(self.JANAF_cachedir)

    def search(self, searchstr):
        """
        List all the species containing a string. Helpful for
        interactive use of the database.
        returns a pandas dataframe containing valid phases.

        >>> db = Janafdb()
        >>> s = db.search('Rb-')
        >>> print(s)
             formula           name phase filename
        1710     Rb-  Rubidium, Ion     g   Rb-007
        >>> s = db.search('Ti')
        >>> print(len(s))
        88
        """

        formulasearch = self.db['formula'].str.contains(searchstr)
        namesearch = self.db['name'].str.contains(searchstr)

        return self.db[formulasearch | namesearch]

    def getphasedata(self, formula=None, name=None, phase=None, nocache=False):
        """
        Returns an element instance given the name of the element.
        formula, name and phase match the respective fields in the JANAF index.
        nocache = True means that we will always get the data from the web.

        >>> db = Janafdb()
        >>> db.getphasedata(formula='O2Ti', phase='cr')
        Traceback (most recent call last):
            ...
        ValueError: There are 2 records matching this pattern.
        >>> db.getphasedata(formula='Oxyz')
        Traceback (most recent call last):
            ...
        ValueError: Valid phase types are ['cr', 'l', 'cr,l', 'g', 'ref', 'cd', 'fl', 'am', 'vit', 'mon', 'pol', 'sln', 'aq', 'sat'].
        >>> db.getphasedata(formula='Oxyz', phase='l')
        Traceback (most recent call last):
            ...
        ValueError: Did not find Oxyz, None, (l)
        """

        # Check that the phase type requested is valid.
        if phase not in Janafdb.VALIDPHASETYPES:
            raise ValueError("Valid phase types are " + str(
                Janafdb.VALIDPHASETYPES) + ".")

        # We can search on either an exact formula, partial text match in the
        # name, and exact phase type.
        formulasearch = pd.Series(np.ones(len(self.db)), dtype=bool)
        namesearch = formulasearch.copy()
        phasesearch = formulasearch.copy()
        if formula is not None:
            formulasearch = self.db['formula'] == formula
        if name is not None:
            namesearch = self.db['name'].str.contains(name)
        if phase is not None:
            phasesearch = self.db['phase'] == phase
        searchmatch = formulasearch & namesearch & phasesearch

        # Get the record (should be one record) which specifies this phase.
        PhaseRecord = self.db[searchmatch]
        if len(PhaseRecord) == 0:
            raise ValueError("Did not find %s, %s, (%s)" %
                             (formula, name, phase))
        if len(PhaseRecord) > 1:
            raise ValueError("There are %d records matching this pattern." %
                             len(PhaseRecord))

        # At this point we have one record.  Check if we have that file cached.
        cachedfilename = os.path.join(
            self.JANAF_cachedir, PhaseRecord['filename'].values[0] + '.txt')
        if os.path.exists(cachedfilename) and not nocache:
            # Yes it was cached, so let's read it into memory.
            with open(cachedfilename, 'r') as f:
                textdata = f.read()
        else:
            # No it was not cached so let's get it from the web.
            response = urllib2.urlopen(Janafdb.JANAF_URL %
                                       PhaseRecord['filename'].values[0])
            textdata = response.read()

            # And cache the data so we aren't making unnecessary trips to the
            # web.
            if not nocache:
                with open(cachedfilename, 'w') as f:
                    f.write(textdata)

        # Create a phase class and return it.
        return JanafPhase(textdata)

    # def getmixturedata(self, components):
    #     """
    #     Creates a mixture of components given a list of tuples
    #     containing the formula and the volume percent
    #     """
    #     mixture = Mixture()
    #     for comp in components:
    #         mixture.add(self.getelementdata(comp[0]), comp[1])
    #
    #     return mixture



from setuptools import setup, find_packages
import thermochem

setup(name="thermochem",
      version=thermochem.__version__,
      description="Some utilities for Thermodynamics and Thermochemistry",
      author="Adel Qalieh",
      author_email="adelq@sas.upenn.edu",
      url="https://github.com/adelq/thermochem",
      license="BSD",
      packages=find_packages(),
      include_package_data=True,
      install_requires=['scipy>=0.6.0', 'numpy>=1.2.1', 'pandas>=0.17.0'],
      zip_safe=False,
      keywords='thermo chemistry physics',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2.5',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Software Development :: Libraries :: Python Modules',
          ],
      )
